# Docs: https://docs.sublime.security/reference/emailrep-introduction

import requests
from threading import Lock
from ..models import Ihunt
from ..stdout import echo

BASE_URL = "https://emailrep.io"

# TODO: The API key is not approved yet on https://emailrep.io/, so I can't implement the function.
# Query: Email
# Return: Info
def req_emailrep_email(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching Emailrep...", ihunt.verbose)

    url = BASE_URL + f"/{ihunt.query.value}"

    headers = {
        "Key": ihunt.apikeys.emailrep,
        "User-Agent": ihunt.user_agent,
    }

    try:
        resp = requests.get(url, headers=headers)
        if resp.status_code == 200:
            with lock:
                d = resp.json()
                if ihunt.data.email is None:
                    ihunt.data.email = d["email"]
                if ihunt.data.domain_exists is None:
                    ihunt.data.domain_exists = d["details"]["domain_exists"]
                if ihunt.data.domain_reputation is None:
                    ihunt.data.domain_reputation = d["details"]["domain_reputation"]
                if ihunt.data.domain_new is None:
                    ihunt.data.domain_new = d["details"]["new_domain"]
                if ihunt.data.free_provider is None:
                    ihunt.data.free_provider = d["details"]["free_provider"]
                if ihunt.data.disposable is None:
                    ihunt.data.disposable = d["details"]["disposable"]
                if ihunt.data.deliverable is None:
                    ihunt.data.deliverable = d["details"]["deliverable"]
                if ihunt.data.spoofable is None:
                    ihunt.data.spoofable = d["details"]["spoofable"]
                if ihunt.data.dmarc_enforced is None:
                    ihunt.data.dmarc_enforced = d["details"]["dmarc_enforced"]
                if ihunt.data.accept_all is None:
                    ihunt.data.accept_all = d["details"]["accept_all"]
                if ihunt.data.spam is None:
                    ihunt.data.spam = d["details"]["spam"]
                if ihunt.data.suspicious is None:
                    ihunt.data.suspicious = d["suspicious"]
                if ihunt.data.blacklisted is None:
                    ihunt.data.blacklisted = d["details"]["blacklisted"]
                if ihunt.data.malicious_activity is None:
                    ihunt.data.malicious_activity = d["details"]["malicious_activity"]
                if ihunt.data.malicious_activity_recent is None:
                    ihunt.data.malicious_activity_recent = d["details"]["malicious_activity_recent"]
                if ihunt.data.credentials_leaked is None:
                    ihunt.data.credentials_leaked = d["details"]["credentials_leaked"]
                if ihunt.data.credentials_leaked_recent is None:
                    ihunt.data.credentials_leaked_recent = d["details"]["credentials_leaked_recent"]
                if ihunt.data.data_breach is None:
                    ihunt.data.data_breach = d["details"]["data_breach"]
    except Exception as e:
        echo(f"[x] Emailrep API error: {e}", ihunt.verbose)

    echo("[*] Finished fetching Emailrep...", ihunt.verbose)


from dataclasses import fields
import ipaddress
import json
from typing import Any


def is_ip_address(s: str) -> bool:
    try:
        ipaddress.ip_address(s)
        return True
    except:
        return False


def has_duplicate(arr: list[Any]) -> bool:
    return len(arr) != len(set(arr))


# The function is used when stdout or write to the output file.
def remove_null_values_in_dict(data: dict[Any]) -> dict[Any]:
    return {
        k: remove_null_values_in_dict(v)
        if isinstance(v, dict)
        else v
        for k, v in data.items()
        if v is not None
    }


# Get field names from dataclass.
def get_class_field_names(Any) -> list[str]:
    return [field.name for field in fields(Any)]


# The function is used for creating prompt for AI-related APIs.
def create_prompt(query: str, data_class: Any) -> str:
    return f"""
Give me some information about the email address "{query}". The rules are below:

- Keep your answers concise.
- The output format is JSON.
- Do not use commas for number values.
- Skip for unknown items.
- Do not use code block.
- The keys of the JSON: {", ".join(get_class_field_names(data_class))}
"""


# Update Data from given JSON object
def update_data_from_json(data: object, json_obj: dict[str, Any]) -> object:
    for field in fields(data):
        current_value = getattr(data, field.name)
        if field.name in json_obj and current_value is None:
            setattr(data, field.name, json_obj[field.name])



from duckduckgo_search import DDGS
import json
from threading import Lock
from ..models import DataEmail, DataOrg, DataPerson, DataUrl, Ihunt
from ..stdout import echo
from ..utils import create_prompt, update_data_from_json


# Query: Email
# Return: Info
def req_duckduckgo_email(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching DuckDuckGo...", ihunt.verbose)
    try:
        with lock:
            results = DDGS().chat(create_prompt(ihunt.query.value, DataEmail), model='claude-3-haiku')
            update_data_from_json(ihunt.data, json.loads(results))
    except Exception as e:
         echo(f"[x] DuckDuckGo API error: {e}", ihunt.verbose)
    echo("[*] Finished fetching DuckDuckGo...", ihunt.verbose)


# Query: Org
# Return: info
def req_duckduckgo_org(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching DuckDuckGo...", ihunt.verbose)
    try:
        with lock:
            results = DDGS().chat(create_prompt(ihunt.query.value, DataOrg), model='claude-3-haiku')
            update_data_from_json(ihunt.data, json.loads(results))
    except Exception as e:
         echo(f"[x] DuckDuckGo API error: {e}", ihunt.verbose)
    echo("[*] Finished fetching DuckDuckGo...", ihunt.verbose)


# Query: Person
# Return: Info
def req_duckduckgo_person(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching DuckDuckGo...", ihunt.verbose)
    try:
        with lock:
            results = DDGS().chat(create_prompt(ihunt.query.value, DataPerson), model='claude-3-haiku')
            update_data_from_json(ihunt.data, json.loads(results))
    except Exception as e:
         echo(f"[x] DuckDuckGo API error: {e}", ihunt.verbose)
    echo("[*] Finished fetching DuckDuckGo...", ihunt.verbose)


# Query: URL
# Return: Info
def req_duckduckgo_url(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching DuckDuckGo...", ihunt.verbose)
    try:
        with lock:
            results = DDGS().chat(create_prompt(ihunt.query.value, DataUrl), model='claude-3-haiku')
            print(results)
            update_data_from_json(ihunt.data, json.loads(results))
    except Exception as e:
         echo(f"[x] DuckDuckGo API error: {e}", ihunt.verbose)
    echo("[*] Finished fetching DuckDuckGo...", ihunt.verbose)



from duckduckgo_search import DDGS
from enum import Enum
import re
# from .apis.duckduckgo import req_duckduckgo_querytype
from .stdout import echo
from .utils import is_ip_address


# Search models: https://huggingface.co/models?pipeline_tag=token-classification
model_ner = "dslim/bert-base-NER"
# model_ner = "dbmdz/bert-large-cased-finetuned-conll03-english"


class QueryType(Enum):
    DOMAIN  = 1
    EMAIL   = 2
    FILE    = 3
    HASH    = 4
    IP      = 5
    ORG     = 6
    PERSON  = 7
    TEL     = 8
    URL     = 9
    UNKNOWN = 10


# The function is used for detecting query type.
def identify_querytype_with_duckduckgo(query: str, verbose: bool) -> QueryType:
    echo("[*] Detecting the query type with DuckDuckGo API...", verbose)
    prompt = f"""
Which of the following types can the text "{query}" be categorized into? Please answer just the category type.

File Path
Hash
Organization
Person
TEL
Unknown
"""
    try:
        results = DDGS().chat(f"{prompt}", model='claude-3-haiku')
        if results.lower() == "file path":
            return QueryType.FILE
        elif results.lower() == "hash":
            return QueryType.HASH
        elif results.lower() == "organization":
            return QueryType.ORG
        elif results.lower() == "person":
            return QueryType.PERSON
        elif results.lower() == "tel":
            return QueryType.TEL
        else:
            return QueryType.UNKNOWN
    except Exception as e:
         echo(f"[x] DuckDuckGo API error: {e}", verbose)
         return QueryType.UNKNOWN


# Uses LLM NER task.
def identify_querytype_with_huggingface(query: str) -> QueryType:
    import torch
    from transformers import pipeline
    device = (
        "cuda"
        if torch.cuda.is_available()
        else "mps"
        if torch.backends.mps.is_available()
        else "cpu"
    )
    ner = pipeline(task="ner", model=model_ner, device=device)
    entities = ner(query)
    
    if len(entities) == 0:
        return QueryType.UNKNOWN

    for entity in entities:
        entity_type = entity['entity']
        if "-ORG" in entity_type:
            return QueryType.ORG
        elif "-PER" in entity_type:
            return QueryType.PERSON
        else:
            return QueryType.UNKNOWN


def identify_querytype(query: str, verbose: bool) -> QueryType:
    echo("[*] Identifying the query type...", verbose)

    # If it was not detected by DuckDuckGo, try programatically resolutions.
    domain_pattern = re.compile(r'^((?!-)[A-Za-z0-9-]{1,63}(?<!-)\.)?(?!-)[A-Za-z0-9-]{1,63}(?<!-)\.(?!-)[A-Za-z0-9-]{1,63}(?<!-)$')
    email_pattern = re.compile(r'^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$')
    url_pattern = re.compile(r'^(https?|ftp|ssh)://[^\s/$.?#].[^\s]*$', re.IGNORECASE)

    if domain_pattern.match(query):
        return QueryType.DOMAIN
    elif email_pattern.match(query):
        return QueryType.EMAIL
    elif is_ip_address(query):
        return QueryType.IP
    elif url_pattern.match(query):
        return QueryType.URL
    
    # Try DuckDuckGo Chat API.
    query_type = identify_querytype_with_duckduckgo(query, verbose)
    if query_type != QueryType.UNKNOWN:
        return query_type

    # Try HuggingFace API.
    return identify_querytype_with_huggingface(query)


class Query:
    value: str
    type: QueryType

    def __init__(self, value: str, verbose: bool) -> None:
        self.value = value
        self.type = identify_querytype(value, verbose)

    def type_str(self) -> str:
        if self.type == QueryType.DOMAIN:
            return "Domain"
        elif self.type == QueryType.EMAIL:
            return "Email"
        elif self.type == QueryType.FILE:
            return "File"
        elif self.type == QueryType.HASH:
            return "Hash"
        elif self.type == QueryType.IP:
            return "IP"
        elif self.type == QueryType.ORG:
            return "Organization"
        elif self.type == QueryType.PERSON:
            return "Person"
        elif self.type == QueryType.TEL:
            return "TEL"
        elif self.type == QueryType.URL:
            return "URL"
        elif self.type == QueryType.UNKNOWN:
            return "Unknown"


